package org.apache.bcel.generic;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;
import java.util.Stack;
import org.apache.bcel.classfile.AnnotationEntry;
import org.apache.bcel.classfile.Annotations;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.ExceptionTable;
import org.apache.bcel.classfile.LineNumber;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.classfile.LocalVariableTypeTable;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.ParameterAnnotationEntry;
import org.apache.bcel.classfile.ParameterAnnotations;
import org.apache.bcel.classfile.RuntimeVisibleParameterAnnotations;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.util.BCELComparator;
import org.apache.commons.lang3.ArrayUtils;

public class MethodGen extends FieldGenOrMethodGen {
   private static BCELComparator bcelComparator = new BCELComparator() {
      public boolean equals(Object o1, Object o2) {
         FieldGenOrMethodGen THIS = (FieldGenOrMethodGen)o1;
         FieldGenOrMethodGen THAT = (FieldGenOrMethodGen)o2;
         return Objects.equals(THIS.getName(), THAT.getName()) && Objects.equals(THIS.getSignature(), THAT.getSignature());
      }

      public int hashCode(Object o) {
         FieldGenOrMethodGen THIS = (FieldGenOrMethodGen)o;
         return THIS.getSignature().hashCode() ^ THIS.getName().hashCode();
      }
   };
   private String className;
   private Type[] argTypes;
   private String[] argNames;
   private int maxLocals;
   private int maxStack;
   private InstructionList il;
   private boolean stripAttributes;
   private LocalVariableTypeTable localVariableTypeTable;
   private final List<LocalVariableGen> variableList;
   private final List<LineNumberGen> lineNumberList;
   private final List<CodeExceptionGen> exceptionList;
   private final List<String> throwsList;
   private final List<Attribute> codeAttrsList;
   private List<AnnotationEntryGen>[] paramAnnotations;
   private boolean hasParameterAnnotations;
   private boolean haveUnpackedParameterAnnotations;
   private List<MethodObserver> observers;

   private static byte[] getByteCodes(Method method) {
      Code code = method.getCode();
      if (code == null) {
         throw new IllegalStateException(String.format("The method '%s' has no code.", method));
      } else {
         return code.getCode();
      }
   }

   public static BCELComparator getComparator() {
      return bcelComparator;
   }

   public static int getMaxStack(ConstantPoolGen cp, InstructionList il, CodeExceptionGen[] et) {
      BranchStack branchTargets = new BranchStack();
      CodeExceptionGen[] var4 = et;
      int maxStackDepth = et.length;

      for(int var6 = 0; var6 < maxStackDepth; ++var6) {
         CodeExceptionGen element = var4[var6];
         InstructionHandle handlerPc = element.getHandlerPC();
         if (handlerPc != null) {
            branchTargets.push(handlerPc, 1);
         }
      }

      int stackDepth = 0;
      maxStackDepth = 0;
      InstructionHandle ih = il.getStart();

      while(ih != null) {
         Instruction instruction = ih.getInstruction();
         short opcode = instruction.getOpcode();
         int delta = instruction.produceStack(cp) - instruction.consumeStack(cp);
         stackDepth += delta;
         if (stackDepth > maxStackDepth) {
            maxStackDepth = stackDepth;
         }

         if (!(instruction instanceof BranchInstruction)) {
            if (opcode == 191 || opcode == 169 || opcode >= 172 && opcode <= 177) {
               ih = null;
            }
         } else {
            BranchInstruction branch = (BranchInstruction)instruction;
            if (instruction instanceof Select) {
               Select select = (Select)branch;
               InstructionHandle[] targets = select.getTargets();
               InstructionHandle[] var13 = targets;
               int var14 = targets.length;

               for(int var15 = 0; var15 < var14; ++var15) {
                  InstructionHandle target = var13[var15];
                  branchTargets.push(target, stackDepth);
               }

               ih = null;
            } else if (!(branch instanceof IfInstruction)) {
               if (opcode == 168 || opcode == 201) {
                  branchTargets.push(ih.getNext(), stackDepth - 1);
               }

               ih = null;
            }

            branchTargets.push(branch.getTarget(), stackDepth);
         }

         if (ih != null) {
            ih = ih.getNext();
         }

         if (ih == null) {
            BranchTarget bt = branchTargets.pop();
            if (bt != null) {
               ih = bt.target;
               stackDepth = bt.stackDepth;
            }
         }
      }

      return maxStackDepth;
   }

   public static void setComparator(BCELComparator comparator) {
      bcelComparator = comparator;
   }

   public MethodGen(int accessFlags, Type returnType, Type[] argTypes, String[] argNames, String methodName, String className, InstructionList il, ConstantPoolGen cp) {
      super(accessFlags);
      this.variableList = new ArrayList();
      this.lineNumberList = new ArrayList();
      this.exceptionList = new ArrayList();
      this.throwsList = new ArrayList();
      this.codeAttrsList = new ArrayList();
      this.setType(returnType);
      this.setArgumentTypes(argTypes);
      this.setArgumentNames(argNames);
      this.setName(methodName);
      this.setClassName(className);
      this.setInstructionList(il);
      this.setConstantPool(cp);
      boolean abstract_ = this.isAbstract() || this.isNative();
      InstructionHandle start = null;
      InstructionHandle end = null;
      if (!abstract_) {
         start = il.getStart();
         if (!this.isStatic() && className != null) {
            this.addLocalVariable("this", ObjectType.getInstance(className), start, (InstructionHandle)end);
         }
      }

      if (argTypes != null) {
         int size = argTypes.length;
         Type[] var13 = argTypes;
         int var14 = argTypes.length;

         for(int var15 = 0; var15 < var14; ++var15) {
            Type argType = var13[var15];
            if (Type.VOID == argType) {
               throw new ClassGenException("'void' is an illegal argument type for a method");
            }
         }

         int i;
         if (argNames != null) {
            if (size != argNames.length) {
               throw new ClassGenException("Mismatch in argument array lengths: " + size + " vs. " + argNames.length);
            }
         } else {
            argNames = new String[size];

            for(i = 0; i < size; ++i) {
               argNames[i] = "arg" + i;
            }

            this.setArgumentNames(argNames);
         }

         if (!abstract_) {
            for(i = 0; i < size; ++i) {
               this.addLocalVariable(argNames[i], argTypes[i], start, (InstructionHandle)end);
            }
         }
      }

   }

   public MethodGen(Method method, String className, ConstantPoolGen cp) {
      this(method.getAccessFlags(), Type.getReturnType(method.getSignature()), Type.getArgumentTypes(method.getSignature()), (String[])null, method.getName(), className, (method.getAccessFlags() & 1280) == 0 ? new InstructionList(getByteCodes(method)) : null, cp);
      Attribute[] attributes = method.getAttributes();
      Attribute[] var5 = attributes;
      int var6 = attributes.length;

      for(int var7 = 0; var7 < var6; ++var7) {
         Attribute attribute = var5[var7];
         Attribute a = attribute;
         if (!(a instanceof Code)) {
            if (a instanceof ExceptionTable) {
               Collections.addAll(this.throwsList, ((ExceptionTable)a).getExceptionNames());
            } else if (a instanceof Annotations) {
               Annotations runtimeAnnotations = (Annotations)a;
               runtimeAnnotations.forEach((element) -> {
                  this.addAnnotationEntry(new AnnotationEntryGen(element, cp, false));
               });
            } else {
               this.addAttribute(a);
            }
         } else {
            Code c = (Code)a;
            this.setMaxStack(c.getMaxStack());
            this.setMaxLocals(c.getMaxLocals());
            CodeException[] ces = c.getExceptionTable();
            int var14;
            if (ces != null) {
               CodeException[] var12 = ces;
               int var13 = ces.length;

               for(var14 = 0; var14 < var13; ++var14) {
                  CodeException ce = var12[var14];
                  int type = ce.getCatchType();
                  ObjectType cType = null;
                  if (type > 0) {
                     String cen = method.getConstantPool().getConstantString(type, (byte)7);
                     cType = ObjectType.getInstance(cen);
                  }

                  int endPc = ce.getEndPC();
                  int length = getByteCodes(method).length;
                  InstructionHandle end;
                  if (length == endPc) {
                     end = this.il.getEnd();
                  } else {
                     end = this.il.findHandle(endPc);
                     end = end.getPrev();
                  }

                  this.addExceptionHandler(this.il.findHandle(ce.getStartPC()), end, this.il.findHandle(ce.getHandlerPC()), cType);
               }
            }

            Attribute[] cAttributes = c.getAttributes();
            Attribute[] var23 = cAttributes;
            var14 = cAttributes.length;

            for(int var24 = 0; var24 < var14; ++var24) {
               Attribute cAttribute = var23[var24];
               a = cAttribute;
               if (a instanceof LineNumberTable) {
                  ((LineNumberTable)a).forEach((l) -> {
                     InstructionHandle ih = this.il.findHandle(l.getStartPC());
                     if (ih != null) {
                        this.addLineNumber(ih, l.getLineNumber());
                     }

                  });
               } else if (a instanceof LocalVariableTable) {
                  this.updateLocalVariableTable((LocalVariableTable)a);
               } else if (a instanceof LocalVariableTypeTable) {
                  this.localVariableTypeTable = (LocalVariableTypeTable)a.copy(cp.getConstantPool());
               } else {
                  this.addCodeAttribute(a);
               }
            }
         }
      }

   }

   public void addAnnotationsAsAttribute(ConstantPoolGen cp) {
      this.addAll(AnnotationEntryGen.getAnnotationAttributes(cp, super.getAnnotationEntries()));
   }

   public void addCodeAttribute(Attribute a) {
      this.codeAttrsList.add(a);
   }

   public void addException(String className) {
      this.throwsList.add(className);
   }

   public CodeExceptionGen addExceptionHandler(InstructionHandle startPc, InstructionHandle endPc, InstructionHandle handlerPc, ObjectType catchType) {
      if (startPc != null && endPc != null && handlerPc != null) {
         CodeExceptionGen c = new CodeExceptionGen(startPc, endPc, handlerPc, catchType);
         this.exceptionList.add(c);
         return c;
      } else {
         throw new ClassGenException("Exception handler target is null instruction");
      }
   }

   public LineNumberGen addLineNumber(InstructionHandle ih, int srcLine) {
      LineNumberGen l = new LineNumberGen(ih, srcLine);
      this.lineNumberList.add(l);
      return l;
   }

   public LocalVariableGen addLocalVariable(String name, Type type, InstructionHandle start, InstructionHandle end) {
      return this.addLocalVariable(name, type, this.maxLocals, start, end);
   }

   public LocalVariableGen addLocalVariable(String name, Type type, int slot, InstructionHandle start, InstructionHandle end) {
      return this.addLocalVariable(name, type, slot, start, end, slot);
   }

   public LocalVariableGen addLocalVariable(String name, Type type, int slot, InstructionHandle start, InstructionHandle end, int origIndex) {
      byte t = type.getType();
      if (t != 16) {
         int add = type.getSize();
         if (slot + add > this.maxLocals) {
            this.maxLocals = slot + add;
         }

         LocalVariableGen l = new LocalVariableGen(slot, name, type, start, end, origIndex);
         int i;
         if ((i = this.variableList.indexOf(l)) >= 0) {
            this.variableList.set(i, l);
         } else {
            this.variableList.add(l);
         }

         return l;
      } else {
         throw new IllegalArgumentException("Can not use " + type + " as type for local variable");
      }
   }

   public void addObserver(MethodObserver o) {
      if (this.observers == null) {
         this.observers = new ArrayList();
      }

      this.observers.add(o);
   }

   public void addParameterAnnotation(int parameterIndex, AnnotationEntryGen annotation) {
      this.ensureExistingParameterAnnotationsUnpacked();
      if (!this.hasParameterAnnotations) {
         List<AnnotationEntryGen>[] parmList = new List[this.argTypes.length];
         this.paramAnnotations = parmList;
         this.hasParameterAnnotations = true;
      }

      List<AnnotationEntryGen> existingAnnotations = this.paramAnnotations[parameterIndex];
      if (existingAnnotations != null) {
         existingAnnotations.add(annotation);
      } else {
         List<AnnotationEntryGen> l = new ArrayList();
         l.add(annotation);
         this.paramAnnotations[parameterIndex] = l;
      }

   }

   public void addParameterAnnotationsAsAttribute(ConstantPoolGen cp) {
      if (this.hasParameterAnnotations) {
         Attribute[] attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.paramAnnotations);
         if (attrs != null) {
            this.addAll(attrs);
         }

      }
   }

   private Attribute[] addRuntimeAnnotationsAsAttribute(ConstantPoolGen cp) {
      Attribute[] attrs = AnnotationEntryGen.getAnnotationAttributes(cp, super.getAnnotationEntries());
      this.addAll(attrs);
      return attrs;
   }

   private Attribute[] addRuntimeParameterAnnotationsAsAttribute(ConstantPoolGen cp) {
      if (!this.hasParameterAnnotations) {
         return Attribute.EMPTY_ARRAY;
      } else {
         Attribute[] attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.paramAnnotations);
         this.addAll(attrs);
         return attrs;
      }
   }

   private void adjustLocalVariableTypeTable(LocalVariableTable lvt) {
      LocalVariable[] lv = lvt.getLocalVariableTable();
      LocalVariable[] var3 = this.localVariableTypeTable.getLocalVariableTypeTable();
      int var4 = var3.length;

      for(int var5 = 0; var5 < var4; ++var5) {
         LocalVariable element = var3[var5];
         LocalVariable[] var7 = lv;
         int var8 = lv.length;

         for(int var9 = 0; var9 < var8; ++var9) {
            LocalVariable l = var7[var9];
            if (element.getName().equals(l.getName()) && element.getIndex() == l.getOrigIndex()) {
               element.setLength(l.getLength());
               element.setStartPC(l.getStartPC());
               element.setIndex(l.getIndex());
               break;
            }
         }
      }

   }

   public MethodGen copy(String className, ConstantPoolGen cp) {
      Method m = ((MethodGen)this.clone()).getMethod();
      MethodGen mg = new MethodGen(m, className, super.getConstantPool());
      if (super.getConstantPool() != cp) {
         mg.setConstantPool(cp);
         mg.getInstructionList().replaceConstantPool(super.getConstantPool(), cp);
      }

      return mg;
   }

   private void ensureExistingParameterAnnotationsUnpacked() {
      if (!this.haveUnpackedParameterAnnotations) {
         Attribute[] attrs = this.getAttributes();
         ParameterAnnotations paramAnnVisAttr = null;
         ParameterAnnotations paramAnnInvisAttr = null;
         Attribute[] var4 = attrs;
         int var5 = attrs.length;

         for(int var6 = 0; var6 < var5; ++var6) {
            Attribute attribute = var4[var6];
            if (attribute instanceof ParameterAnnotations) {
               if (!this.hasParameterAnnotations) {
                  List<AnnotationEntryGen>[] parmList = new List[this.argTypes.length];
                  this.paramAnnotations = parmList;
                  Arrays.setAll(this.paramAnnotations, (i) -> {
                     return new ArrayList();
                  });
               }

               this.hasParameterAnnotations = true;
               ParameterAnnotations rpa = (ParameterAnnotations)attribute;
               if (rpa instanceof RuntimeVisibleParameterAnnotations) {
                  paramAnnVisAttr = rpa;
               } else {
                  paramAnnInvisAttr = rpa;
               }

               ParameterAnnotationEntry[] parameterAnnotationEntries = rpa.getParameterAnnotationEntries();

               for(int j = 0; j < parameterAnnotationEntries.length; ++j) {
                  ParameterAnnotationEntry immutableArray = rpa.getParameterAnnotationEntries()[j];
                  List<AnnotationEntryGen> mutable = this.makeMutableVersion(immutableArray.getAnnotationEntries());
                  this.paramAnnotations[j].addAll(mutable);
               }
            }
         }

         if (paramAnnVisAttr != null) {
            this.removeAttribute(paramAnnVisAttr);
         }

         if (paramAnnInvisAttr != null) {
            this.removeAttribute(paramAnnInvisAttr);
         }

         this.haveUnpackedParameterAnnotations = true;
      }
   }

   public boolean equals(Object obj) {
      return bcelComparator.equals(this, obj);
   }

   public List<AnnotationEntryGen> getAnnotationsOnParameter(int i) {
      this.ensureExistingParameterAnnotationsUnpacked();
      return this.hasParameterAnnotations && i <= this.argTypes.length ? this.paramAnnotations[i] : null;
   }

   public String getArgumentName(int i) {
      return this.argNames[i];
   }

   public String[] getArgumentNames() {
      return (String[])this.argNames.clone();
   }

   public Type getArgumentType(int i) {
      return this.argTypes[i];
   }

   public Type[] getArgumentTypes() {
      return (Type[])this.argTypes.clone();
   }

   public String getClassName() {
      return this.className;
   }

   public Attribute[] getCodeAttributes() {
      return (Attribute[])this.codeAttrsList.toArray(Attribute.EMPTY_ARRAY);
   }

   private CodeException[] getCodeExceptions() {
      int size = this.exceptionList.size();
      CodeException[] cExc = new CodeException[size];
      Arrays.setAll(cExc, (i) -> {
         return ((CodeExceptionGen)this.exceptionList.get(i)).getCodeException(super.getConstantPool());
      });
      return cExc;
   }

   public CodeExceptionGen[] getExceptionHandlers() {
      return (CodeExceptionGen[])this.exceptionList.toArray(CodeExceptionGen.EMPTY_ARRAY);
   }

   public String[] getExceptions() {
      return (String[])this.throwsList.toArray(ArrayUtils.EMPTY_STRING_ARRAY);
   }

   private ExceptionTable getExceptionTable(ConstantPoolGen cp) {
      int size = this.throwsList.size();
      int[] ex = new int[size];
      Arrays.setAll(ex, (i) -> {
         return cp.addClass((String)this.throwsList.get(i));
      });
      return new ExceptionTable(cp.addUtf8("Exceptions"), 2 + 2 * size, ex, cp.getConstantPool());
   }

   public InstructionList getInstructionList() {
      return this.il;
   }

   public LineNumberGen[] getLineNumbers() {
      return (LineNumberGen[])this.lineNumberList.toArray(LineNumberGen.EMPTY_ARRAY);
   }

   public LineNumberTable getLineNumberTable(ConstantPoolGen cp) {
      int size = this.lineNumberList.size();
      LineNumber[] ln = new LineNumber[size];
      Arrays.setAll(ln, (i) -> {
         return ((LineNumberGen)this.lineNumberList.get(i)).getLineNumber();
      });
      return new LineNumberTable(cp.addUtf8("LineNumberTable"), 2 + ln.length * 4, ln, cp.getConstantPool());
   }

   public LocalVariableGen[] getLocalVariables() {
      int size = this.variableList.size();
      LocalVariableGen[] lg = new LocalVariableGen[size];
      this.variableList.toArray(lg);

      for(int i = 0; i < size; ++i) {
         if (lg[i].getStart() == null && this.il != null) {
            lg[i].setStart(this.il.getStart());
         }

         if (lg[i].getEnd() == null && this.il != null) {
            lg[i].setEnd(this.il.getEnd());
         }
      }

      if (size > 1) {
         Arrays.sort(lg, Comparator.comparingInt(LocalVariableGen::getIndex));
      }

      return lg;
   }

   public LocalVariableTable getLocalVariableTable(ConstantPoolGen cp) {
      LocalVariableGen[] lg = this.getLocalVariables();
      int size = lg.length;
      LocalVariable[] lv = new LocalVariable[size];
      Arrays.setAll(lv, (i) -> {
         return lg[i].getLocalVariable(cp);
      });
      return new LocalVariableTable(cp.addUtf8("LocalVariableTable"), 2 + lv.length * 10, lv, cp.getConstantPool());
   }

   public LocalVariableTypeTable getLocalVariableTypeTable() {
      return this.localVariableTypeTable;
   }

   public int getMaxLocals() {
      return this.maxLocals;
   }

   public int getMaxStack() {
      return this.maxStack;
   }

   public Method getMethod() {
      String signature = this.getSignature();
      ConstantPoolGen cp = super.getConstantPool();
      int nameIndex = cp.addUtf8(super.getName());
      int signatureIndex = cp.addUtf8(signature);
      byte[] byteCode = this.il != null ? this.il.getByteCode() : null;
      LineNumberTable lnt = null;
      LocalVariableTable lvt = null;
      if (!this.variableList.isEmpty() && !this.stripAttributes) {
         this.updateLocalVariableTable(this.getLocalVariableTable(cp));
         this.addCodeAttribute(lvt = this.getLocalVariableTable(cp));
      }

      if (this.localVariableTypeTable != null) {
         if (lvt != null) {
            this.adjustLocalVariableTypeTable(lvt);
         }

         this.addCodeAttribute(this.localVariableTypeTable);
      }

      if (!this.lineNumberList.isEmpty() && !this.stripAttributes) {
         this.addCodeAttribute(lnt = this.getLineNumberTable(cp));
      }

      Attribute[] codeAttrs = this.getCodeAttributes();
      int attrsLen = 0;
      Attribute[] var10 = codeAttrs;
      int excLen = codeAttrs.length;

      for(int var12 = 0; var12 < excLen; ++var12) {
         Attribute codeAttr = var10[var12];
         attrsLen += codeAttr.getLength() + 6;
      }

      CodeException[] cExc = this.getCodeExceptions();
      excLen = cExc.length * 8;
      Code code = null;
      Attribute[] parameterAnnotations;
      Attribute[] annotations;
      if (byteCode != null && !this.isAbstract() && !this.isNative()) {
         annotations = this.getAttributes();
         parameterAnnotations = annotations;
         int var15 = annotations.length;

         for(int var16 = 0; var16 < var15; ++var16) {
            Attribute a = parameterAnnotations[var16];
            if (a instanceof Code) {
               this.removeAttribute(a);
            }
         }

         code = new Code(cp.addUtf8("Code"), 8 + byteCode.length + 2 + excLen + 2 + attrsLen, this.maxStack, this.maxLocals, byteCode, cExc, codeAttrs, cp.getConstantPool());
         this.addAttribute(code);
      }

      annotations = this.addRuntimeAnnotationsAsAttribute(cp);
      parameterAnnotations = this.addRuntimeParameterAnnotationsAsAttribute(cp);
      ExceptionTable et = null;
      if (!this.throwsList.isEmpty()) {
         this.addAttribute(et = this.getExceptionTable(cp));
      }

      Method m = new Method(super.getAccessFlags(), nameIndex, signatureIndex, this.getAttributes(), cp.getConstantPool());
      if (lvt != null) {
         this.removeCodeAttribute(lvt);
      }

      if (this.localVariableTypeTable != null) {
         this.removeCodeAttribute(this.localVariableTypeTable);
      }

      if (lnt != null) {
         this.removeCodeAttribute(lnt);
      }

      if (code != null) {
         this.removeAttribute(code);
      }

      if (et != null) {
         this.removeAttribute(et);
      }

      this.removeRuntimeAttributes(annotations);
      this.removeRuntimeAttributes(parameterAnnotations);
      return m;
   }

   public Type getReturnType() {
      return this.getType();
   }

   public String getSignature() {
      return Type.getMethodSignature(super.getType(), this.argTypes);
   }

   public int hashCode() {
      return bcelComparator.hashCode(this);
   }

   private List<AnnotationEntryGen> makeMutableVersion(AnnotationEntry[] mutableArray) {
      List<AnnotationEntryGen> result = new ArrayList();
      AnnotationEntry[] var3 = mutableArray;
      int var4 = mutableArray.length;

      for(int var5 = 0; var5 < var4; ++var5) {
         AnnotationEntry element = var3[var5];
         result.add(new AnnotationEntryGen(element, this.getConstantPool(), false));
      }

      return result;
   }

   public void removeCodeAttribute(Attribute a) {
      this.codeAttrsList.remove(a);
   }

   public void removeCodeAttributes() {
      this.localVariableTypeTable = null;
      this.codeAttrsList.clear();
   }

   public void removeException(String c) {
      this.throwsList.remove(c);
   }

   public void removeExceptionHandler(CodeExceptionGen c) {
      this.exceptionList.remove(c);
   }

   public void removeExceptionHandlers() {
      this.exceptionList.clear();
   }

   public void removeExceptions() {
      this.throwsList.clear();
   }

   public void removeLineNumber(LineNumberGen l) {
      this.lineNumberList.remove(l);
   }

   public void removeLineNumbers() {
      this.lineNumberList.clear();
   }

   public void removeLocalVariable(LocalVariableGen l) {
      l.dispose();
      this.variableList.remove(l);
   }

   public void removeLocalVariables() {
      this.variableList.forEach(LocalVariableGen::dispose);
      this.variableList.clear();
   }

   public void removeLocalVariableTypeTable() {
      this.localVariableTypeTable = null;
   }

   public void removeNOPs() {
      InstructionHandle next;
      if (this.il != null) {
         for(InstructionHandle ih = this.il.getStart(); ih != null; ih = next) {
            next = ih.getNext();
            if (next != null && ih.getInstruction() instanceof NOP) {
               try {
                  this.il.delete(ih);
               } catch (TargetLostException var12) {
                  TargetLostException e = var12;
                  InstructionHandle[] var4 = e.getTargets();
                  int var5 = var4.length;

                  for(int var6 = 0; var6 < var5; ++var6) {
                     InstructionHandle target = var4[var6];
                     InstructionTargeter[] var8 = target.getTargeters();
                     int var9 = var8.length;

                     for(int var10 = 0; var10 < var9; ++var10) {
                        InstructionTargeter targeter = var8[var10];
                        targeter.updateTarget(target, next);
                     }
                  }
               }
            }
         }
      }

   }

   public void removeObserver(MethodObserver o) {
      if (this.observers != null) {
         this.observers.remove(o);
      }

   }

   public void removeRuntimeAttributes(Attribute[] attrs) {
      Attribute[] var2 = attrs;
      int var3 = attrs.length;

      for(int var4 = 0; var4 < var3; ++var4) {
         Attribute attr = var2[var4];
         this.removeAttribute(attr);
      }

   }

   public void setArgumentName(int i, String name) {
      this.argNames[i] = name;
   }

   public void setArgumentNames(String[] argNames) {
      this.argNames = argNames;
   }

   public void setArgumentType(int i, Type type) {
      this.argTypes[i] = type;
   }

   public void setArgumentTypes(Type[] argTypes) {
      this.argTypes = argTypes;
   }

   public void setClassName(String className) {
      this.className = className;
   }

   public void setInstructionList(InstructionList il) {
      this.il = il;
   }

   public void setMaxLocals() {
      if (this.il != null) {
         int max = this.isStatic() ? 0 : 1;
         int index;
         if (this.argTypes != null) {
            Type[] var2 = this.argTypes;
            int var3 = var2.length;

            for(index = 0; index < var3; ++index) {
               Type argType = var2[index];
               max += argType.getSize();
            }
         }

         for(InstructionHandle ih = this.il.getStart(); ih != null; ih = ih.getNext()) {
            Instruction ins = ih.getInstruction();
            if (ins instanceof LocalVariableInstruction || ins instanceof RET || ins instanceof IINC) {
               index = ((IndexedInstruction)ins).getIndex() + ((TypedInstruction)ins).getType(super.getConstantPool()).getSize();
               if (index > max) {
                  max = index;
               }
            }
         }

         this.maxLocals = max;
      } else {
         this.maxLocals = 0;
      }

   }

   public void setMaxLocals(int m) {
      this.maxLocals = m;
   }

   public void setMaxStack() {
      if (this.il != null) {
         this.maxStack = getMaxStack(super.getConstantPool(), this.il, this.getExceptionHandlers());
      } else {
         this.maxStack = 0;
      }

   }

   public void setMaxStack(int m) {
      this.maxStack = m;
   }

   public void setReturnType(Type returnType) {
      this.setType(returnType);
   }

   public void stripAttributes(boolean flag) {
      this.stripAttributes = flag;
   }

   public final String toString() {
      String access = Utility.accessToString(super.getAccessFlags());
      String signature = Type.getMethodSignature(super.getType(), this.argTypes);
      signature = Utility.methodSignatureToString(signature, super.getName(), access, true, this.getLocalVariableTable(super.getConstantPool()));
      StringBuilder buf = new StringBuilder(signature);
      Attribute[] var4 = this.getAttributes();
      int var5 = var4.length;

      for(int var6 = 0; var6 < var5; ++var6) {
         Attribute a = var4[var6];
         if (!(a instanceof Code) && !(a instanceof ExceptionTable)) {
            buf.append(" [").append(a).append("]");
         }
      }

      if (!this.throwsList.isEmpty()) {
         Iterator var8 = this.throwsList.iterator();

         while(var8.hasNext()) {
            String throwsDescriptor = (String)var8.next();
            buf.append("\n\t\tthrows ").append(throwsDescriptor);
         }
      }

      return buf.toString();
   }

   public void update() {
      if (this.observers != null) {
         Iterator var1 = this.observers.iterator();

         while(var1.hasNext()) {
            MethodObserver observer = (MethodObserver)var1.next();
            observer.notify(this);
         }
      }

   }

   private void updateLocalVariableTable(LocalVariableTable a) {
      this.removeLocalVariables();
      LocalVariable[] var2 = a.getLocalVariableTable();
      int var3 = var2.length;

      for(int var4 = 0; var4 < var3; ++var4) {
         LocalVariable l = var2[var4];
         InstructionHandle start = this.il.findHandle(l.getStartPC());
         InstructionHandle end = this.il.findHandle(l.getStartPC() + l.getLength());
         if (null == start) {
            start = this.il.getStart();
         }

         this.addLocalVariable(l.getName(), Type.getType(l.getSignature()), l.getIndex(), start, end, l.getOrigIndex());
      }

   }

   static final class BranchTarget {
      final InstructionHandle target;
      final int stackDepth;

      BranchTarget(InstructionHandle target, int stackDepth) {
         this.target = target;
         this.stackDepth = stackDepth;
      }
   }

   static final class BranchStack {
      private final Stack<BranchTarget> branchTargets = new Stack();
      private final Hashtable<InstructionHandle, BranchTarget> visitedTargets = new Hashtable();

      public BranchTarget pop() {
         return !this.branchTargets.empty() ? (BranchTarget)this.branchTargets.pop() : null;
      }

      public void push(InstructionHandle target, int stackDepth) {
         if (!this.visited(target)) {
            this.branchTargets.push(this.visit(target, stackDepth));
         }
      }

      private BranchTarget visit(InstructionHandle target, int stackDepth) {
         BranchTarget bt = new BranchTarget(target, stackDepth);
         this.visitedTargets.put(target, bt);
         return bt;
      }

      private boolean visited(InstructionHandle target) {
         return this.visitedTargets.get(target) != null;
      }
   }
}
